"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.DocumentContext = void 0;
const assert_1 = __importDefault(require("assert"));
const events_1 = require("events");
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
/**
 * @internal
 */
class DocumentContext extends events_1.EventEmitter {
    constructor(routingKey, head, log, getLatestTail, getContextManagerPauseState) {
        super();
        this.routingKey = routingKey;
        this.log = log;
        this.getLatestTail = getLatestTail;
        this.getContextManagerPauseState = getContextManagerPauseState;
        this.closed = false;
        this.contextError = undefined;
        // Below flag is used to track whether head has been updated after a pause/resume event.
        // This is to allow moving out of order once during resume.
        // Value = true means it is in a paused state and waiting to be updated during resume.
        this.headPaused = false;
        // Head represents the largest offset related to the document that is not checkpointed.
        // Tail will be set to the checkpoint offset of the previous head
        this.headInternal = head;
        this.tailInternal = this.getLatestTail();
        this.lastSuccessfulOffsetInternal = this.tailInternal.offset; // will be -1 at creation
    }
    get head() {
        return this.headInternal;
    }
    get tail() {
        return this.tailInternal;
    }
    get lastSuccessfulOffset() {
        return this.lastSuccessfulOffsetInternal;
    }
    /**
     * Returns whether or not there is pending work in flight - i.e. the head and tail are not equal
     */
    hasPendingWork() {
        return this.headInternal !== this.tailInternal;
    }
    /**
     * Sets the last successfully processed offset.
     */
    setLastSuccessfulOffset(offset) {
        this.lastSuccessfulOffsetInternal = offset;
    }
    /**
     * Sets the state to pause, i.e. headPaused = true, without emitting the pause event.
     * It is different than pause() method which emits the pause event.
     * This is used to set the state to pause when another doc in the same kafka partition triggered pause and we want to pause all the docs in that kafka partition.
     */
    setStateToPause() {
        this.headPaused = true;
    }
    /**
     * Updates the head offset for the context.
     */
    setHead(head) {
        (0, assert_1.default)(head.offset > this.head.offset || this.headPaused, `Head offset ${head.offset} must be greater than the current head offset ${this.head.offset} or headPaused should be true (${this.headPaused}). Topic ${head.topic}, partition ${head.partition}, tenantId ${this.routingKey.tenantId}, documentId ${this.routingKey.documentId}.`);
        // If head is moving backwards
        if (head.offset <= this.head.offset) {
            if (head.offset <= this.tail.offset) {
                server_services_telemetry_1.Lumberjack.info("Not updating documentContext head since new head's offset is <= last checkpoint offset (tail), returning early", {
                    newHeadOffset: head.offset,
                    currentHeadOffset: this.head.offset,
                    currentTailOffset: this.tail.offset,
                    documentId: this.routingKey.documentId,
                });
                return false;
            }
            // allow moving backwards
            server_services_telemetry_1.Lumberjack.info("Allowing the document context head to move backwards to the specified offset", {
                newHeadOffset: head.offset,
                currentHeadOffset: this.head.offset,
                currentTailOffset: this.tail.offset,
                headPaused: this.headPaused,
                documentId: this.routingKey.documentId,
            });
        }
        // When moving back to a state where head and tail differ we set the tail to be the old head, as in the
        // constructor, to make tail represent the inclusive top end of the checkpoint range.
        if (!this.hasPendingWork()) {
            this.tailInternal = this.getLatestTail();
        }
        if (this.headPaused) {
            server_services_telemetry_1.Lumberjack.info("Setting headPaused to false", {
                newHeadOffset: head.offset,
                currentHeadOffset: this.head.offset,
                currentTailOffset: this.tail.offset,
                documentId: this.routingKey.documentId,
            });
            this.headPaused = false;
        }
        this.headInternal = head;
        return true;
    }
    checkpoint(message, restartOnCheckpointFailure) {
        if (this.closed) {
            return;
        }
        // skip checkpoint in paused state
        const contextManagerPauseState = this.getContextManagerPauseState();
        if (this.headPaused ||
            (!contextManagerPauseState.headPaused && contextManagerPauseState.tailPaused)) {
            const telemetryProperties = {
                documentId: this.routingKey.documentId,
                tenantId: this.routingKey.tenantId,
                headOffset: this.head.offset,
                tailOffset: this.tail.offset,
                checkpointOffset: message.offset,
                headPaused: this.headPaused,
            };
            if (this.headPaused) {
                server_services_telemetry_1.Lumberjack.info("Skipping doc checkpoint since the documentContext is currently in paused state", telemetryProperties);
            }
            else {
                // contextManager's tail is resumed after head, so its possible to be in this state, but vice versa is not possible
                // but document shouldnt be checkpointing at this time, adding a log here to monitor if this happens
                server_services_telemetry_1.Lumberjack.warning("Skipping doc checkpoint since contextManager's tail is not yet updated after resume", telemetryProperties);
            }
            return;
        }
        // Assert offset is between the current tail and head
        const offset = message.offset;
        (0, assert_1.default)(offset > this.tail.offset && offset <= this.head.offset, `Checkpoint offset ${offset} must be greater than the current tail offset ${this.tail.offset} and less than or equal to the head offset ${this.head.offset}. Topic ${message.topic}, partition ${message.partition}, tenantId ${this.routingKey.tenantId}, documentId ${this.routingKey.documentId}.`);
        // Update the tail and broadcast the checkpoint
        this.tailInternal = message;
        this.emit("checkpoint", restartOnCheckpointFailure);
    }
    error(error, errorData) {
        if (this.closed) {
            // don't emit errors after closing
            server_services_telemetry_1.Lumberjack.info("Skipping emitting error since the documentContext is already closed", {
                documentId: this.routingKey.documentId,
                tenantId: this.routingKey.tenantId,
            });
            return;
        }
        this.contextError = error;
        server_services_telemetry_1.Lumberjack.verbose("Emitting error from documentContext");
        this.emit("error", error, errorData);
    }
    close() {
        this.closed = true;
        this.removeAllListeners();
    }
    getContextError() {
        return this.contextError;
    }
    pause(offset, reason) {
        this.headPaused = true;
        this.emit("pause", offset, reason);
    }
    resume() {
        this.emit("resume");
    }
}
exports.DocumentContext = DocumentContext;
//# sourceMappingURL=documentContext.js.map